/*****************************************************************************
File:    dploader.c
Purpose: Reset Hostess i/S2, load a binary file into the dual port RAM,
         signal the COM processor to begin execution or to start Turbo
         Debugger.
Company: Comtrol Corporation
Author:  Craig Harrison
Comment: This version of DPLOADER allows the control program to be placed in
         controller memory anywhere after 0:0c7F.  Defines ENTRY_SEG and
         ENTRY_OFF in DPRAM.H determine where the control program will be
         placed. Macro CPC_RELOC_SEG in the make file must match ENTRY_SEG.

         Usage: dploader [control program name] [Turbo Debugger (y/n)]
         where:
            "control program name" is an optional name of the control program
               file to download.  If not given this defaults to "cpc.bin".
            "Turbo Debugger (y/n)" is an optional "y" or "n" switch indicating
               whether or not to invoke Turbo Debugger (R) remotely to allow
               on board debugging of the control program.
Ident:   $Id$
*******************************************************************************
Copyright 1994 Comtrol Corporation. All rights reserved.
Subject to developers license agreement.
*******************************************************************************
*/
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <dos.h>
#include "dpram.h"
#include "hilib.h"

FIRMUSER_T far *fu_p;         /* ptr to firmuser user area */
int io;                       /* controller base I/O address */
char cpfname[13];             /* name of control pgm binary file to download */
int do_tdrem;                 /* set if Turbo Debugger Remote to be invoked */

/* CR3 values for various window sizes and offsets */
char w16k[8] = {0x38,0x39,0x3a,0x3b,0x3c,0x3d,0x3e,0x3f};  /* 16K windows */
char w32k[4] = {0x30,0x32,0x34,0x36};            /* 32K windows */
char w64k[2] = {0x20,0x24};                      /* 64K windows */

/* CR4 values for IRQ 0-15, 0 means IRQ is disabled */
char cr4[16] = {0,0,0,0x08,0x09,0x0a,0,0,0,0x0b,0x0c,0x0d,0x0e,0,0,0x0f};

void cr_init(void);
void download(void);
void reset_board(void);
int invoke_tdrem(void);
void dly(int msec);


main(int argc,char *argv[])
{
   if(argc < 2)                        /* if no cmd line args use defaults */
   {
      strcpy(cpfname,"cpc.bin");
      do_tdrem = 0;
   }
   else                                /* get cmd line args */
   {
      if(argv[1][0] == '?' || argc != 3)
      {
         printf("\nUsage: dploader [control program name] [Turbo Debugger (y/n)]\n");
         exit(0);
      }
      else
      {
         strncpy(cpfname,argv[1],13);  /* get control pgm name */
         cpfname[12] = 0;              /* make sure it's null terminated */
         if(argv[2][0] == 'Y' || argv[2][0] == 'y')
            do_tdrem = 1;
         else
            do_tdrem = 0;
      }
   }

   fu_p = (FIRMUSER_T far *)FIRMUSER_ADDR_SYS; /* firmuser area address */
   io = IO_SYS;                        /* I/O base address */
   reset_board();                      /* reset board if needed */
   download();                         /* download & start ctrlpgm */
   return(0);
}

/*****************************************************************************
Purpose: Reset board, initialize control registers.
Entry:   Nothing
Returns: Nothing
*/
void reset_board(void)
{
   int i;
   unsigned rec_flag;

   outp(io+3,0);
   dly(100);
   outp(io+3,0xff);
   printf("Waiting for reset to complete\n");
   dly(500);
   cr_init();
   for(i = 0;i < 21;i++)
   {
      dly(1000);
      if(fu_p->i_flag == 0x55aa)
      break;
   }
   if(((rec_flag = fu_p->i_flag) != 0x55aa) && i >= 21)
   {
      printf("   Reset failed, interaction flag = %XH\n",rec_flag);
      exit(1);
   }
}

/*****************************************************************************
Purpose: Load a binary file into dual port memory using the firmware
         utilities, signal the COM Processor to start execution,
         verify execution started.
Entry:   Nothing
Returns: Nothing
Notes:   The file to be downloaded must have an entry point of
         ENTRY_SEG:ENTRY_OFF and must write 55aaH to the interaction flag.

         This download uses the firmware utilities, which require a source
         buffer in DPRAM (dual port RAM).  This source buffer is 256 bytes
         in size and is located at 0:c00, which is where the trace buffer
         will be once the control program starts.  After the download
         finishes this area is zero'd out so it won't be confusing if the
         trace buffer is used later.
*/
void download(void)
{
   char far *ram_p;
   unsigned wait;                      /* Do nothing loop counter */
   unsigned rec_flag;                  /* Received copy of interact flag*/
   unsigned long absaddr;              /* Download absolute address */
   static char buf[256];               /* Buffer for data to be xfered
                                          to dual port RAM */
   long int nbytes;                    /* Number of bytes downloaded */
   int f_h,                            /* File handle */
       bytesread;                      /* Bytes read on current read */

   /* Check interaction flag */
   if((rec_flag = fu_p->i_flag) != 0x55aa)
   {
      printf("Bad interaction flag = %XH before download\n",rec_flag);
      exit(1);
   }

   /* Open control program file for read */
   if((f_h = open(cpfname,O_BINARY)) == -1)
   {
      printf("Unable to open control program file %s.\n",cpfname);
      exit(1);
   }

   /* Set up source buf address in firmware user area for copy command */
   absaddr = ((unsigned long)(ENTRY_SEG) << 4) + ENTRY_OFF;
   fu_p->cmd = COPY;            /* firmware utility copy command */
   fu_p->msg[0] = (char)((long)LOAD_ADDR_CTL >> 16); /* source buffer segment */
   fu_p->msg[1] = (char)((long)LOAD_ADDR_CTL >> 24);
   fu_p->msg[2] = (char)LOAD_ADDR_CTL;       /* source buffer offset */
   fu_p->msg[3] = (char)((long)LOAD_ADDR_CTL >> 8);

   /* Download */
   nbytes = 0;
   while(!eof(f_h))
   {
      ram_p = (char far *)LOAD_ADDR_SYS; /* Source buf at ctrlpgm entry pnt */
      bytesread = read(f_h,buf,256);   /* Read 256 bytes into buffer */
      if(bytesread <= -1)              /* Error reading file */
      {
         printf("   Error reading file %s\n",cpfname);
         exit(1);
      }
      nbytes += bytesread;
      _fmemcpy(ram_p,buf,bytesread);   /* Move buf into DPRAM source buf */
      ram_p += bytesread;

      /* Firmware utility copy of DPRAM source buf bytes to destination */
      fu_p->status = 0;             /* Clear status */
      fu_p->msg[4] = (char)(absaddr >> 4);  /* Destination segment */
      fu_p->msg[5] = (char)(absaddr >> 12);
      fu_p->msg[6] = (char)(absaddr & 0x0000000fL);  /* Destination offset */
      fu_p->msg[7] = 0;
      fu_p->msg[8] = (char)bytesread;
      fu_p->msg[9] = (char)(bytesread >> 8);
      outp(io + 2,0);                  /* Interrupt COM Processor */
      for(wait = 0;wait < 10;wait++)   /* Wait for copy to finish */
      {
         if(fu_p->status == DONE)
            break;
         dly(2);
      }
      if((wait >= 10) && (fu_p->status != DONE)) /* Copy failed */
      {
         printf("Firmware copy command failed\n");
         exit(1);
      }
      absaddr += bytesread;            /* Next destination address */
   }

   /* Download successful, start control program by interrupting COM Processor
      using firmware utility */
   close(f_h);                         /* close ctrlpgm file */
   printf("%ld bytes downloaded successfully\n",nbytes);
   if(invoke_tdrem())                  /* Turbo Debugger invoked */
      return;                          /* don't start ctrlpgm */
   _fmemset((void far *)LOAD_ADDR_SYS,0,256); /* Zero out source buf area */
   fu_p->i_flag = 0xaa55;              /* set up interaction flag */
   fu_p->cmd = EXEC;             /* firmware utility command to exec ctrlpgm */
   fu_p->status = 0;
   fu_p->msg[0] = (unsigned)ENTRY_SEG;   /* entry point segment */
   fu_p->msg[1] = (unsigned)ENTRY_SEG >> 8;
   fu_p->msg[2] = (unsigned)ENTRY_OFF;   /* entry point offset */
   fu_p->msg[3] = (unsigned)ENTRY_OFF >> 8;
   outp(io + 2,0);                     /* Interrupt COM Processor */
   printf("COM processor interrupted to start control program\n");
   for(wait = 0;wait < 10;wait++)      /* wait for control program to */
   {                                   /*   restore interaction flag */
      dly(400);
      if(fu_p->i_flag == 0x55aa)
         break;
   }
   if((wait >= 10) && ((rec_flag = fu_p->i_flag) != 0x55aa))
   {
      printf("Control program failed to start, interact flag = %XH\n",
              rec_flag);
      exit(1);
   }
   else
   {
      printf("Control program started execution\n");
   }
}

/*****************************************************************************
Purpose: HOSTESS i control register initialization
Entry:   Nothing
Returns: Nothing
*/
void cr_init(void)
{
   outp(io+1,CR1 | DIS_RAM | AT_MODE);      /* access CR1, AT mode */
   outp(io,0);                              /* below 1 Meg */
   outp(io+1,CR2 | DIS_RAM);                /* access CR2 */
   outp(io,CR2_VAL);                        /* below 1 Meg RAM address */
   outp(io+1,CR3 | DIS_RAM);                /* access CR3 */
   outp(io,w64k[0]);                        /* 64K lower window */
   outp(io+1,CR4 | DIS_RAM);                /* access CR4 */
   outp(io,cr4[0]);                         /* disable IRQ */
   outp(io+1,EN_RAM);                       /* enable DPRAM */
}

/*****************************************************************************
Purpose: Invoke Turbo Debugger Remote kernel, if needed.
Entry:   Nothing
Returns: 1 if tdrem is invoked, else 0
Comment: We will stuff the int 27H instruction to invoke the TDREM kernel into
         the download buffer area and execute there.
*/
int invoke_tdrem()
{
   unsigned save_word;                 /* Saved copy of overwritten word */
   unsigned far *inst_p;               /* Ptr to control pgm entry point */

   if(do_tdrem)
   {
      inst_p = (unsigned far *)LOAD_ADDR_SYS; /* Ptr to int 27H instruc addr*/
      save_word = *inst_p;             /* Save word to be overwritten */
      *inst_p = 0x27cd;                /* Write int 27H instruction used to
                                          invoke TDREM kernel */
      fu_p->cmd = EXEC;           /* firmware utility command to exec ctrlpgm */
      fu_p->status = 0;
      fu_p->msg[0] = (char)((long)LOAD_ADDR_CTL >> 16); /* instruction addr segment */
      fu_p->msg[1] = (char)((long)LOAD_ADDR_CTL >> 24);
      fu_p->msg[2] = (char)LOAD_ADDR_CTL;       /* instruction addr offset */
      fu_p->msg[3] = (char)((long)LOAD_ADDR_CTL >> 8);
      outp(io + 2,0);                  /* Interrupt COM Processor */
      dly(10);                         /* Give it a chance to execute */
      printf("Turbo Debugger Remote Kernel Started\n");
      *inst_p = save_word;             /* Restore overwritten word */
   }
   return(do_tdrem);
}

/*****************************************************************************
Purpose: Delay for given number of milliseconds
Entry:   Number of milliseconds to delay
Returns: Nothing
*/
void dly(int msec)
{
   long i,t;

   t = (long)msec * 500L;              /* each inp() takes 2uS max */
   for(i = 0;i < t;i++)
   {
      inp(io);
   }
}

